
//		Sift.cpp - file sorting plugin for TCC


#define UNICODE					1
#define _UNICODE				1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR				0
#define VER_MINOR				71
#define VER_BUILD				2
#define VER_PATCH				2

#define PLUGIN_NAME				L"Sift"

#ifdef _WIN64
#define PLATFORM				L"x64"
#else
#define PLATFORM				L"x86"
#endif


#define MAX_ARG_LEN				8192
#define SHORT_BUF_LEN			1024
#define MAX_FILENAME_LEN		16384
#define MAX_ARGS				1024
#define MAX_ITEMS				4096

#define ERRHANDLE				GetStdHandle( STD_ERROR_HANDLE )


LPTSTR arg[MAX_ARGS];
int arglen[MAX_ARGS];
int numargs = 0;
wchar_t allargs[MAX_ARG_LEN]	= L"";

int SiftFound = -1;

#define MODE_UNKNOWN			0
#define MODE_LARGEST			1
#define MODE_SMALLEST			2
#define MODE_OLDEST				3
#define MODE_NEWEST				4

#define STAMP_LAST_WRITE		0
#define STAMP_CREATED			1
#define STAMP_LAST_ACCESS		2

#define SIFT_RECURSIVE			1
#define SIFT_NO_JUNCTIONS		2
#define SIFT_NO_HIDDEN_DIRS		4


#define BATON_START				2048
#define BATON_RESET				128

#define FAKE_SEMICOLON			0xff1b

#define MIN_USER_WIDTH			80
#define MAX_USER_WIDTH			1024

#include "Ranges.cpp"


const wchar_t PluginFeatures[] = L"sift,_sift,@sift";


const wchar_t HelpWindowTitle[] = L"Sift Plugin Help";


const wchar_t SiftHelpText[] = L"Find the n largest (smallest, etc.) files and sort them into an array.\n\
\n\
SIFT /A: /C:sep /I /R /T: /U /V /Z n criterion array filespec...\r\n\
\n\
\t/A:  attributes mask\n\
\t/C:  time separator character\n\
\t/I   date stamps in ISO 8601 format, YYYY-MM-DD\n\
\t/R   recurse into subdirectories\n\
\t/T:  date stamp to use; default is W\n\
\t/U   date stamps as UTC, not localized\n\
\t/V   dump array to stdout\n\
\t/Z   localize date stamps dumbly, like DIR\n\
\n\
\tn    number of files to return; defaults to 10\n\
\tcriterion is LARGEST, SMALLEST, OLDEST, NEWEST; or /L /S /O /N  \n\
\n\
An n-by-3 array will be created to hold the resulting list of files.\n";


const wchar_t _SiftHelpText[] = L"_SIFT\n\n\
Returns the number of files found by the last SIFT command.  \n";


const wchar_t f_SiftHelpText[] = L"@SIFT[filespec,criterion,attribs,flags,stamp]\n\n\
criterion is L, S, O, or N for Largest, Smallest, Oldest, Newest\n\
flags is bitmapped:  1 recurse, 2 no junctions, 4 no hidden dirs  \n\
stamp is A, C, or W for last Access, Creation, or last Write\n\n\
Returns the full name of the single largest (smallest, etc.)\n\
matching file.\n";



struct FileEntry {
	FileEntry *Next, *Prev;
	__int64 FileSize, FileStamp;
	BOOL VolumeFAT, VolumeNTFS;
	wchar_t Filename;
} ;



FileEntry *FirstItem = NULL, *LastItem = NULL;
int NumItems = 0;


wchar_t ModeName[5][10] = { L"unknown", L"largest", L"smallest", L"oldest", L"newest" };
wchar_t StampName[3][12] = { L"Last Write", L"Created", L"Last Access" };
wchar_t BoolValue[2][4] = { L"No", L"Yes" };


const wchar_t BatonChar[5] = L"/-\\|";
int BatonCount = BATON_START;
int BatonIndex = 0;
int ScreenWidth = 0;



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( GetStdHandle(STD_ERROR_HANDLE), shortbuf );

	return 0;
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



int UnsupportedVersion( void )
{
	wchar_t TempBuffer[SHORT_BUF_LEN] = L"%_4ver";
	wchar_t *sep1 = NULL;
	long int ShellMajor = -1, ShellMinor = -1, ShellBuild = -1;

	ExpandVariables( TempBuffer, 1 );
	ShellMajor = wcstol( TempBuffer, &sep1, 10 );

	if ( *sep1 ) {
		sep1++;
		if ( *sep1 )
			ShellMinor = wcstol( sep1, NULL, 10 );
	}

	wcscpy_s( TempBuffer, SHORT_BUF_LEN, L"%_build" );
	ExpandVariables( TempBuffer, 1 );
	ShellBuild = wcstol( TempBuffer, NULL, 10 );

	if ( ( ShellMajor < 7 ) || ( ShellMajor > 99 ) || ( ShellMinor < 0 ) || ( ShellMinor > 99 ) || ( ShellBuild < 0 ) || ( ShellBuild > 9999 ) )
		return -1;			//  version number fails sanity checks

	if ( ShellMajor < 10 )
		return 1;

	wcscpy_s( TempBuffer, SHORT_BUF_LEN, L"%_cmdproc" );
	ExpandVariables( TempBuffer, 1 );
	if ( !_wcsicmp( TempBuffer, L"TCCLE" ) )
		return 2;

	return 0;
}



int ParseArgs( LPTSTR inString, int flags )
{
	/*
		flags:  bit 0 - break args at spaces
				bit 1 - break args at commas (either bit 0 or bit 1 must be set)
				bit 2 - slashes kludge:   treat /a/b like /a /b
				bit 3 - quotes kludge:    treat /a"foo" like /a:"foo"
				bit 4 - equals kludge:    break at first unquoted equals sign
				bit 5 - one-arg kludge:   if arg does not begin with /, allow unquoted spaces
				bit 7 - force uppercase
	*/

	if ( !inString )
		return -1;

	numargs = 0;

	int argnum = 0;
	int i = 0;
	int a = 0;
	int b = 0;
	wchar_t ch = '\0';
	BOOL quote  = FALSE;
	BOOL adv    = FALSE;
	BOOL done   = FALSE;
	int eqkl    = 0;

	if ( ( flags & 3 ) == 0 )
		flags |= 1;

	for ( i = 0; i < MAX_ARGS; i++ ) {
		arg[i] = NULL;
		arglen[i] = 0;
	}

	for ( i = 0; !done; i++ ) {
		ch = inString[i];
		adv = FALSE;

		if ( ch == '"' ) {
			quote = !quote;

			if ( quote && ( a == 2 ) && ( arg[argnum][0] == '/' ) && ( iswalpha( arg[argnum][1] ) ) && ( flags & 8 ) )
				ch = ':';
			else
				continue;
		}

		if ( iswspace( ch ) && ( a == 0 ) )
			continue;

		if ( ( a == 0 ) && ( ch != '/' ) && ( ( flags & 33 ) == 33 ) )
			flags &= 0xfffffe;

		if ( iswspace( ch ) && !quote && ( a == 1 ) && ( argnum > 0 ) && ( flags & 16 ) && ( arg[argnum][0] == '=' ) ) {
			continue;
		}

		if ( iswspace( ch ) && !quote && ( flags & 1 ) )
			adv = TRUE;

		if ( ( ch == ',' ) && !quote && ( flags & 2 ) )
			adv = TRUE;

		if ( ( ch == '/' ) && !quote && ( a > 1 ) && ( arg[argnum][0] == '/' ) && ( flags & 4 ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
		}

		if ( ( ch == '=' ) && !quote && ( eqkl == 0 ) && ( a > 0 ) && ( flags & 16 ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			eqkl = argnum;
			flags &= 0xfffffe;
			//- Printf( L"   * kludge invoked; eqkl = %d\n", eqkl );
		}
		else if ( ( ch == '=' ) && !quote && ( flags & 16 ) ) {
			eqkl = -1;
			flags &= 0xfffffe;
		}

		if ( ch == '\0' ) {
			if ( a )
				argnum++;

			done = TRUE;
			continue;
		}

		if ( adv ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			continue;
		}

		if ( b >= ( MAX_ARG_LEN - 1 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow: b = %d, MAX_ARG_LEN = %d \r\n", b, MAX_ARG_LEN );
			return -1;
		}

		if ( argnum < MAX_ARGS ) {
			if ( a == 0 )
				arg[argnum] = allargs + b;

			if ( flags & 128 )
				ch = towupper( ch );

			a++;
			arglen[argnum]  = a;
			allargs[b++] = ch;
			allargs[b]   = '\0';
		}

	}

	for ( i = 0; i < argnum; i++ ) {
		while ( ( arglen[i] > 0 ) && iswspace( arg[i][arglen[i] - 1] ) ) {
			arglen[i]--;
			arg[i][arglen[i]] = '\0';
		}
	}

	numargs = argnum;
	return argnum;
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned long int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned decimal

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar = sep1;

		return 0;
	}

	return -1;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



int InsertCommas( LPTSTR Buffer, unsigned int BufferLen )
{
	if ( !Buffer )
		return -1;

	if ( wcslen( Buffer ) > ( MAX_ARG_LEN - 10 ) )
		return 1;

	wchar_t temp[MAX_ARG_LEN]	= L"%@comma[";
	wcscat_s( temp, MAX_ARG_LEN, Buffer );
	wcscat_s( temp, MAX_ARG_LEN, L"]" );

	if ( ExpandVariables( temp, 1 ) )
		return 2;

	if ( wcslen( temp ) >= BufferLen )
		return 3;

	wcscpy_s( Buffer, BufferLen, temp );
	return 0;
}



int CallInternal( LPTSTR Cmd, LPTSTR Args )
{
	if ( !Cmd )
		return -1;
	else if ( Cmd[0] == '\0' )
		return -1;


	wchar_t CmdBuf[MAX_ARG_LEN]		= L"";
	wchar_t ExpVar[SHORT_BUF_LEN]	= L"%_expansion";
	BOOL AddStar					= TRUE;


	ExpandVariables( ExpVar, 1 );
	if ( wcschr( ExpVar, '1' ) )
		AddStar = FALSE;

	if ( ( Cmd[0] != '*' ) && AddStar )
		wcscpy_s( CmdBuf, MAX_ARG_LEN, L"*" );

	wcscat_s( CmdBuf, MAX_ARG_LEN, Cmd );

	if ( Args ) {
		if ( Args[0] ) {
			if ( !iswspace( Args[0] ) )
				wcscat_s( CmdBuf, MAX_ARG_LEN, L" " );

			wcscat_s( CmdBuf, MAX_ARG_LEN, Args );
		}
	}

	return Command( CmdBuf, 0 );
}



BOOL IsBetter( __int64 FoundItemSize, __int64 FoundItemStamp, FileEntry Test, int Mode )
{

	switch ( Mode ) {

		case MODE_SMALLEST:
			if ( FoundItemSize < Test.FileSize )
				return TRUE;
			break;

		case MODE_OLDEST:
			if ( FoundItemStamp < Test.FileStamp )
				return TRUE;
			break;

		case MODE_NEWEST:
			if ( FoundItemStamp > Test.FileStamp )
				return TRUE;
			break;

		default:
			if ( FoundItemSize > Test.FileSize )
				return TRUE;
	}

	return FALSE;
}



int GetVolInfo( LPTSTR inString, LPTSTR outFSType )
{
	if ( !inString || !outFSType )
		return -1;

	wchar_t tempbuf[MAX_ARG_LEN] = L"";
	wchar_t volname[SHORT_BUF_LEN] = L"";
	int numBackslashes = 4;
	int i = 0, j = 0, k = 0;
	wchar_t ch = '\0';

	if ( iswalpha(inString[0]) && ( inString[1] == ':' ) && ( inString[2] == '\\' ) )
		numBackslashes = 1;
	else if ( ( inString[0] != '\\' ) || ( inString[1] != '\\' ) )
		return -1;

	do {
		ch = inString[i++];

		if ( k == numBackslashes )
			ch = '\0';

		if ( ch == '\\' )
			k++;

		tempbuf[j++] = ch;
	} while ( ch != '\0' );

	outFSType[0] = '\0';

	if ( !GetVolumeInformation( tempbuf, volname, SHORT_BUF_LEN, NULL, NULL, NULL, outFSType, MAX_ARG_LEN ) )
		return -1;

	if ( outFSType[0] == '\0' )
		return -1;

	return 0;
}



int StampToString( LPTSTR outString, __int64 inTimeI, BOOL SuppressTime, BOOL BeSmart, BOOL UTC, BOOL ISOFormat, LPTSTR TimeSepString )
{
	if ( !outString || !TimeSepString )
		return -1;

	FILETIME inTime;
	SYSTEMTIME FileStamp, FileStampLocal;
	FILETIME FileTimeLocal;
	wchar_t shortbuf[SHORT_BUF_LEN] = L"";

	inTime.dwHighDateTime = ( inTimeI >> 32 );
	inTime.dwLowDateTime  = ( inTimeI & 0xffffffff );

	if ( UTC )
		FileTimeToSystemTime( &inTime, &FileStampLocal );
	else if ( BeSmart ) {
		FileTimeToSystemTime( &inTime, &FileStamp );
		SystemTimeToTzSpecificLocalTime( NULL, &FileStamp, &FileStampLocal );
	}
	else {
		FileTimeToLocalFileTime( &inTime, &FileTimeLocal );
		FileTimeToSystemTime( &FileTimeLocal, &FileStampLocal );
	}


	if ( ISOFormat ) {
		if ( SuppressTime )
			Sprintf( outString, L"%04d-%02d-%02d", FileStampLocal.wYear, FileStampLocal.wMonth, FileStampLocal.wDay );
		else
			Sprintf( outString, L"%04d-%02d-%02d%s%02d:%02d:%02d", FileStampLocal.wYear, FileStampLocal.wMonth, FileStampLocal.wDay, TimeSepString, FileStampLocal.wHour, FileStampLocal.wMinute, FileStampLocal.wSecond );
	}
	else {
		GetDateFormat( LOCALE_USER_DEFAULT, DATE_LONGDATE, &FileStampLocal, NULL, outString, SHORT_BUF_LEN );

		if ( !SuppressTime ) {
			wcscat_s( outString, MAX_ARG_LEN, TimeSepString );
			GetTimeFormat( LOCALE_USER_DEFAULT, 0, &FileStampLocal, NULL, shortbuf, SHORT_BUF_LEN );
			wcscat_s( outString, MAX_ARG_LEN, shortbuf );
		}
	}

	if ( UTC )
		wcscat_s( outString, MAX_ARG_LEN, L" UTC" );

	return 0;
}



__int64 DirectorySize( LPTSTR DirName )
{
	HANDLE SearchHandle = NULL;
	WIN32_FIND_DATA FoundItem;
	__int64 Sum = 0;

	SearchHandle = FindFirstFile( DirName, &FoundItem );
	if ( SearchHandle == INVALID_HANDLE_VALUE )
		return 0;

	do {
		if ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
			Sum += ( (__int64) FoundItem.nFileSizeHigh << 32 ) | FoundItem.nFileSizeLow;

	} while ( FindNextFile( SearchHandle, &FoundItem ) );

	FindClose( SearchHandle );
	return Sum;
}



int HideSemicolons( LPTSTR inFilename )
{
	if ( !inFilename )
		return -1;
	else if ( inFilename[0] == '\0' )
		return 0;

	int rv = 0;

	for ( int i = 0; inFilename[i]; i++ )
		if ( inFilename[i] == ';' ) {
			inFilename[i] = FAKE_SEMICOLON;
			rv++;
		}

	return rv;
}



int RestoreSemicolons( LPTSTR inFilename )
{
	if ( !inFilename )
		return -1;
	else if ( inFilename[0] == '\0' )
		return 0;

	int rv = 0;

	for ( int i = 0; inFilename[i]; i++ )
		if ( inFilename[i] == FAKE_SEMICOLON ) {
			inFilename[i] = ';';
			rv++;
		}

	return rv;
}



int WildcardComparison2( LPTSTR Wildspec, LPTSTR Filename, int Extension, int Brackets )
{
	if ( !Wildspec || !Filename )
		return -1;
	else if ( ( Wildspec[0] == '\0' ) || ( Filename[0] == '\0' ) )
		return -1;

	int j = 0;
	wchar_t Temp[MAX_ARG_LEN] = L"";
	wchar_t ch = 0x01;

	for ( int i = 0; ch; i++ ) {
		ch = Wildspec[i];
		if ( ( ch == ';' ) || ( ch == '\0' ) ) {
			if ( j )
				if ( !WildcardComparison( Temp, Filename, Extension, Brackets ) )
					return 0;

			j = 0;
			Temp[j] = '\0';
		}
		else if ( j < ( MAX_ARG_LEN - 1 ) ) {
			Temp[j++] = ch;
			Temp[j] = '\0';
		}
	}

	return 1;
}



FileEntry *CreateEntry( LPTSTR Filename, __int64 FileSize, __int64 FileStamp, BOOL VolFAT, BOOL VolNTFS )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;

	int len = (int) wcslen( Filename );

	FileEntry *ptr = (FileEntry*) AllocMem( sizeof( FileEntry ) + ( ( len + 1 ) * sizeof( wchar_t ) ) );
	if ( !ptr )
		return NULL;

	wcscpy_s( &(ptr->Filename), len + 1, Filename );
	ptr->FileSize   = FileSize;
	ptr->FileStamp  = FileStamp;
	ptr->VolumeFAT  = VolFAT;
	ptr->VolumeNTFS = VolNTFS;
	ptr->Next       = NULL;
	ptr->Prev       = NULL;

	return ptr;
}



int TrimTail( int n )
{
	FileEntry *Prev = NULL;

	while ( NumItems > n ) {
		Prev = LastItem->Prev;
		Prev->Next = NULL;
		FreeMem( LastItem );
		LastItem = Prev;
		NumItems--;
	}

	return 0;
}



int FreeList( void )
{
	FileEntry *This		= FirstItem;
	FileEntry *Next		= NULL;

	while ( This ) {
		Next = This->Next;
		FreeMem( This );
		This = Next;
	}

	FirstItem = LastItem = NULL;
	return 0;
}



int SiftFiles( LPTSTR DirName, LPTSTR WildSpec, DWORD AttribMask, DWORD AttribValues, int n, int Mode, int WhichStamp, int Recurse, BOOL FullPaths, LPTSTR ExcludeDescriptions, void *FirstRange, BOOL Debug, BOOL VolumeIsFAT, BOOL VolumeIsNTFS )
{
	wchar_t FullName[MAX_FILENAME_LEN] = L"";
	wcscpy_s( FullName, MAX_FILENAME_LEN, DirName );
	wcscat_s( FullName, MAX_FILENAME_LEN, L"*" );

	if ( Debug )
		Printf( L"   * SiftFiles( \"%s\", \"%s\" )\r\n", DirName, WildSpec );

	HANDLE SearchHandle = NULL;
	WIN32_FIND_DATA FoundItem;
	__int64 FoundItemSize = 0, FoundItemStamp = 0;

	SearchHandle = FindFirstFile( FullName, &FoundItem );
	if ( SearchHandle == INVALID_HANDLE_VALUE )
		return 0;

	FileEntry *NewItem = NULL, *Search = NULL, *Prev = NULL;


	do {
		if ( FoundItem.cFileName[0] == '.' )
			if ( ( FoundItem.cFileName[1] == '\0' ) || ( ( FoundItem.cFileName[1] == '.' ) && ( FoundItem.cFileName[2] == '\0' ) ) )
				continue;

		if ( ScreenWidth ) {
			BatonCount--;
			if ( BatonCount < 0 ) {
				_putwch( BatonChar[BatonIndex] );
				_putwch( 0x08 );
				BatonCount = BATON_RESET;
				BatonIndex = ( BatonIndex + 1 ) & 0x03;
			}
		}

		if ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
			if ( ( Recurse & SIFT_RECURSIVE ) && ( ( ( Recurse & SIFT_NO_JUNCTIONS ) == 0 ) || ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT ) == 0 ) ) && ( ( ( Recurse & SIFT_NO_HIDDEN_DIRS ) == 0 ) || ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN ) == 0 ) ) ) {
				wcscpy_s( FullName, MAX_FILENAME_LEN, DirName );
				wcscat_s( FullName, MAX_FILENAME_LEN, FoundItem.cFileName );
				wcscat_s( FullName, MAX_FILENAME_LEN, L"\\" );
				SiftFiles( FullName, WildSpec, AttribMask, AttribValues, n, Mode, WhichStamp, Recurse, FullPaths, ExcludeDescriptions, FirstRange, Debug, VolumeIsFAT, VolumeIsNTFS );
			}

		if ( ( FoundItem.dwFileAttributes & AttribMask ) != ( AttribValues & AttribMask ) )
			continue;

		if ( WildcardComparison2( WildSpec, FoundItem.cFileName, 1, 1 ) )
			continue;

		if ( ExcludeDescriptions )
			if ( ExcludeDescriptions[0] )
				if ( ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 ) && !_wcsicmp( FoundItem.cFileName, ExcludeDescriptions ) )
					continue;

		if ( CheckRanges( &FoundItem, DirName, FirstRange ) )
			continue;

		if ( FullPaths )
			wcscpy_s( FullName, MAX_FILENAME_LEN, DirName );
		else
			FullName[0] = '\0';

		wcscat_s( FullName, MAX_FILENAME_LEN, FoundItem.cFileName );

		if ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) {
			wchar_t Name[MAX_FILENAME_LEN] = L"";
			wcscpy_s( Name, MAX_FILENAME_LEN, DirName );
			wcscat_s( Name, MAX_FILENAME_LEN, FoundItem.cFileName );
			wcscat_s( Name, MAX_FILENAME_LEN, L"\\*" );
			FoundItemSize = DirectorySize( Name );
		}
		else
			FoundItemSize  = ( (__int64) FoundItem.nFileSizeHigh << 32 ) | FoundItem.nFileSizeLow;

		switch ( WhichStamp ) {
			case STAMP_CREATED:
				FoundItemStamp = ( (__int64) FoundItem.ftCreationTime.dwHighDateTime << 32 ) | FoundItem.ftCreationTime.dwLowDateTime;
				break;

			case STAMP_LAST_ACCESS:
				FoundItemStamp = ( (__int64) FoundItem.ftLastAccessTime.dwHighDateTime << 32 ) | FoundItem.ftLastAccessTime.dwLowDateTime;
				break;

			default:
				FoundItemStamp = ( (__int64) FoundItem.ftLastWriteTime.dwHighDateTime << 32 ) | FoundItem.ftLastWriteTime.dwLowDateTime;
		}

		NewItem = CreateEntry( FullName, FoundItemSize, FoundItemStamp, VolumeIsFAT, VolumeIsNTFS );
		//***  check for error!

		if ( !FirstItem ) {
			FirstItem = NewItem;
			LastItem  = NewItem;
			NumItems = 1;
		}
		else if ( IsBetter( FoundItemSize, FoundItemStamp, *FirstItem, Mode ) ) {
			NewItem->Next = FirstItem;
			FirstItem->Prev = NewItem;
			FirstItem = NewItem;

			NumItems++;
			if ( NumItems > n )
				TrimTail( n );
		}
		else if ( !IsBetter( FoundItemSize, FoundItemStamp, *LastItem, Mode ) ) {
			if ( NumItems < n ) {
				NewItem->Prev = LastItem;
				LastItem->Next = NewItem;
				LastItem = NewItem;
				NumItems++;
			}
			else
				FreeMem( NewItem );
		}
		else {
			Search = FirstItem;
			Prev   = NULL;

			while ( !IsBetter( FoundItemSize, FoundItemStamp, *Search, Mode ) ) {
				Prev = Search;
				Search = Search->Next;
			}

			Prev->Next    = NewItem;
			NewItem->Prev = Prev;
			Search->Prev  = NewItem;
			NewItem->Next = Search;

			NumItems++;
			if ( NumItems > n )
				TrimTail( n );
		}

	} while ( FindNextFile( SearchHandle, &FoundItem ) );


	FindClose( SearchHandle );
	return 0;
}



int SiftFiles2( LPTSTR Filespec, DWORD AttribMask, DWORD AttribValues, int n, int Mode, int WhichStamp, int Recurse, BOOL FullPaths, LPTSTR ExcludeDescriptions, void *FirstRange, BOOL Debug )
{
	if ( !Filespec )
		return -1;
	else if ( Filespec[0] == '\0' )
		return -1;


	wchar_t DirName[MAX_FILENAME_LEN]  = L"";
	wchar_t WildSpec[MAX_FILENAME_LEN] = L"";

	wchar_t FSType[SHORT_BUF_LEN] = L"";
	BOOL VolumeIsFAT = FALSE, VolumeIsNTFS = FALSE;


	//  call QueryTrueName() to canonicalize paths, resolve directory aliases, etc.:

	BOOL DoQTN = TRUE;
	if ( iswalpha( Filespec[0] ) )
		if ( Filespec[1] == ':' )
			if ( Filespec[2] == '\0' )
				if ( !QueryDriveReady( towupper( Filespec[0] ) - 'A' + 1 ) )
					DoQTN = FALSE;

	if ( DoQTN ) {
		HideSemicolons( Filespec );
		QueryTrueName( Filespec, DirName );
		RestoreSemicolons( DirName );
	}

	if ( DirName[0] == '\0' )
		wcscpy_s( DirName, MAX_FILENAME_LEN, Filespec );


	//  get filesystem info for this directory:

	if ( !GetVolInfo( DirName, FSType ) ) {
		VolumeIsFAT  = !_wcsnicmp( FSType, L"FAT", 3 );
		VolumeIsNTFS = !_wcsnicmp( FSType, L"NTFS", 4 );
	}


	//  supply asterisk wildspec if needed:

	int lastbs = 0;
	for ( int i = 0; DirName[i]; i++ ) {
		if ( DirName[i] == '\\' )
			lastbs = ( i + 1 );

		if ( ( i == 1 ) && ( DirName[i] == ':' ) && iswalpha( DirName[0] ) )
			lastbs = 2;
	}

	if ( QueryIsDirectory( DirName ) ) {
		wcscpy_s( WildSpec, MAX_FILENAME_LEN, L"*" );

		if ( DirName[ wcslen( DirName ) - 1 ] != '\\' )
			wcscat_s( DirName, MAX_FILENAME_LEN, L"\\" );
	}
	else if ( lastbs ) {
		if ( DirName[lastbs] )
			wcscpy_s( WildSpec, MAX_FILENAME_LEN, DirName + lastbs );
		else
			wcscpy_s( WildSpec, MAX_FILENAME_LEN, L"*" );

		DirName[lastbs] = '\0';
	}
	else {
		wcscpy_s( WildSpec, MAX_FILENAME_LEN, DirName );
		wcscpy_s( DirName, MAX_FILENAME_LEN, L"%_cwds" );
		ExpandVariables( DirName, 1 );
	}


	// print debug info if desired:

	if ( Debug ) {
		Printf( L"   * SiftFiles2( \"%s\" )\r\n", Filespec );
		Printf( L"DirName   : \"%s\"\r\n", DirName );
		Printf( L"WildSpec  : \"%s\"\r\n", WildSpec );
	}

	return SiftFiles( DirName, WildSpec, AttribMask, AttribValues, n, Mode, WhichStamp, Recurse, FullPaths, ExcludeDescriptions, FirstRange, Debug, VolumeIsFAT, VolumeIsNTFS );
}



int OutputWidth( )
{
	if ( !QueryIsConsole( GetStdHandle( STD_OUTPUT_HANDLE ) ) )
		return 0;

	HANDLE ConsoleHandle = CreateFile( L"CONOUT$", ( GENERIC_READ | GENERIC_WRITE ), ( FILE_SHARE_READ | FILE_SHARE_WRITE ), 0, OPEN_EXISTING, NULL, NULL );
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	int Width = 0, rv = 0;

	if ( ConsoleHandle != INVALID_HANDLE_VALUE ) {
		rv = GetConsoleScreenBufferInfo( ConsoleHandle, &ConsoleInfo );
		CloseHandle( ConsoleHandle );

		if ( rv )
			if ( ( ConsoleInfo.dwSize.X >= 40 ) && ( ConsoleInfo.dwSize.X <= 1024 ) )
				Width = ConsoleInfo.dwSize.X;
	}

	return Width;
}



int Report( int Found, DWORD AttribMask, DWORD AttribValues, int Mode, LPTSTR ArrayName, int ArrayRows )
{
	wchar_t ItemName[SHORT_BUF_LEN]   = L"";
	wchar_t NumString[SHORT_BUF_LEN]  = L"";

	if ( AttribMask & AttribValues & FILE_ATTRIBUTE_DIRECTORY ) {
		if ( Found == 1 )
			wcscpy_s( ItemName, SHORT_BUF_LEN, L"directory" );
		else
			wcscpy_s( ItemName, SHORT_BUF_LEN, L"directories" );
	}
	else if ( AttribMask & FILE_ATTRIBUTE_DIRECTORY ) {
		if ( Found == 1 )
			wcscpy_s( ItemName, SHORT_BUF_LEN, L"file" );
		else
			wcscpy_s( ItemName, SHORT_BUF_LEN, L"files" );
	}
	else {
		if ( Found == 1 )
			wcscpy_s( ItemName, SHORT_BUF_LEN, L"item" );
		else
			wcscpy_s( ItemName, SHORT_BUF_LEN, L"items" );
	}

	if ( !Found ) {
		Printf( L"No matching %s found.\n", ItemName );
		return 0;
	}

	wsprintf( NumString, L"%d", Found );
	InsertCommas( NumString, SHORT_BUF_LEN );			//	AddCommas( NumString );
	Printf( L"%s %s %s sifted into %s[%d,3].\n", NumString, ModeName[Mode], ItemName, ArrayName, ArrayRows );
	return 0;
}



int RepeatChar( int n, wchar_t ch )
{
	if ( n < 1 )
		return 0;

	wchar_t temp[MAX_ARG_LEN] = L"";

	for ( int i = 0; i < n; i++ )
		temp[i] = ch;

	temp[n] = '\0';
	QPuts( temp );
	return 0;
}



int DumpArray( LPTSTR ArrayName, size_t LongestFilename, size_t LongestSize, size_t LongestStamp, int WhichStamp, int Width )
{
	if ( NumItems == 0 )
		return -1;

	if ( Width <= 0 )
		Width = ( ScreenWidth ) ? ScreenWidth : 100;

	int IndexWidth = 1, SizeWidth = (int) LongestSize, StampWidth = (int) LongestStamp;

	wchar_t Filename[MAX_ARG_LEN]   = L"";
	wchar_t SizeStr[SHORT_BUF_LEN]  = L"";
	wchar_t StampStr[SHORT_BUF_LEN] = L"";
	int len = 0;

	
	if ( NumItems > 1000 )
		IndexWidth = 4;
	else if ( NumItems > 100 )
		IndexWidth = 3;
	else if ( NumItems > 10 )
		IndexWidth = 2;

	if ( SizeWidth < 4 )
		SizeWidth = 4;

	len = (int) wcslen( StampName[WhichStamp] );
	if ( StampWidth < len )
		StampWidth = len;

	int FilenameWidth = Width - SizeWidth - StampWidth - IndexWidth - 7;
	
	if ( FilenameWidth < 30 )
		FilenameWidth = 30;

	if ( FilenameWidth > (int) LongestFilename ) {
		FilenameWidth = (int) LongestFilename;
		if ( FilenameWidth < 10 )
			FilenameWidth = 10;
	}


	Printf( L" \r\n%*s  %-*s  %*s  %*s\r\n", IndexWidth, L"", FilenameWidth, L"Filename", SizeWidth, L"Size", StampWidth, StampName[WhichStamp] );
	RepeatChar( IndexWidth, 0x20 );
	QPuts( L"  " );
	RepeatChar( FilenameWidth, '-' );
	QPuts( L"  " );
	RepeatChar( SizeWidth, '-' );
	QPuts( L"  " );
	RepeatChar( StampWidth, '-' );
	CrLf();

	for ( int i = 0; i < NumItems; i++ ) {
		Printf( L"%*u  ", IndexWidth, i );

		swprintf( Filename, MAX_ARG_LEN, L"%%%s[%d,0]", ArrayName, i );
		ExpandVariables( Filename, 1 );
		len = (int) wcslen( Filename );
		if ( len <= FilenameWidth )
			Printf( L"%-*s  ", FilenameWidth, Filename );
		else
			Printf( L"...%s  ", Filename + ( len - FilenameWidth + 3 ) );

		swprintf( SizeStr, SHORT_BUF_LEN, L"%%%s[%d,1]", ArrayName, i );
		ExpandVariables( SizeStr, 1 );

		swprintf( StampStr, SHORT_BUF_LEN, L"%%%s[%d,2]", ArrayName, i );
		ExpandVariables( StampStr, 1 );

		Printf( L"%*s  %*s\r\n", SizeWidth, SizeStr, StampWidth, StampStr );
	}

	CrLf();
	return 0;
}



int ParseAttribs( LPTSTR inString, DWORD *AttribMask, DWORD *AttribValues )
{
	if ( !inString )
		return -1;

	if ( inString[0] == '\0' ) {
		*AttribMask		= 0;
		*AttribValues	= 0;
		return 0;
	}

	BOOL Plus = TRUE;
	int j = 0;


	while ( inString[j] ) {
		switch ( towupper( inString[j++] ) ) {

			case '+' :
				if ( inString[j] == '\0' )
					return 1;

				Plus = TRUE;
				break;

			case '-' :
				if ( inString[j] == '\0' )
					return 1;

				Plus = FALSE;
				break;

			case 'A' :
				*AttribMask |= FILE_ATTRIBUTE_ARCHIVE;
				if ( Plus )
					*AttribValues |= FILE_ATTRIBUTE_ARCHIVE;
				else
					*AttribValues &= ~FILE_ATTRIBUTE_ARCHIVE;
				break;

			case 'C' :
				*AttribMask |= FILE_ATTRIBUTE_COMPRESSED;
				if ( Plus )
					*AttribValues |= FILE_ATTRIBUTE_COMPRESSED;
				else
					*AttribValues &= ~FILE_ATTRIBUTE_COMPRESSED;
				break;

			case 'D' :
				*AttribMask |= FILE_ATTRIBUTE_DIRECTORY;
				if ( Plus )
					*AttribValues |= FILE_ATTRIBUTE_DIRECTORY;
				else
					*AttribValues &= ~FILE_ATTRIBUTE_DIRECTORY;
				break;

			case 'E' :
				*AttribMask |= FILE_ATTRIBUTE_ENCRYPTED;
				if ( Plus )
					*AttribValues |= FILE_ATTRIBUTE_ENCRYPTED;
				else
					*AttribValues &= ~FILE_ATTRIBUTE_ENCRYPTED;
				break;

			case 'H' :
				*AttribMask |= FILE_ATTRIBUTE_HIDDEN;
				if ( Plus )
					*AttribValues |= FILE_ATTRIBUTE_HIDDEN;
				else
					*AttribValues &= ~FILE_ATTRIBUTE_HIDDEN;
				break;

			case 'I' :
				*AttribMask |= FILE_ATTRIBUTE_NOT_CONTENT_INDEXED;
				if ( Plus )
					*AttribValues |= FILE_ATTRIBUTE_NOT_CONTENT_INDEXED;
				else
					*AttribValues &= ~FILE_ATTRIBUTE_NOT_CONTENT_INDEXED;
				break;

			case 'O' :
				*AttribMask |= FILE_ATTRIBUTE_OFFLINE;
				if ( Plus )
					*AttribValues |= FILE_ATTRIBUTE_OFFLINE;
				else
					*AttribValues &= ~FILE_ATTRIBUTE_OFFLINE;
				break;

			case 'R' :
				*AttribMask |= FILE_ATTRIBUTE_READONLY;
				if ( Plus )
					*AttribValues |= FILE_ATTRIBUTE_READONLY;
				else
					*AttribValues &= ~FILE_ATTRIBUTE_READONLY;
				break;

			case 'S' :
				*AttribMask |= FILE_ATTRIBUTE_SYSTEM;
				if ( Plus )
					*AttribValues |= FILE_ATTRIBUTE_SYSTEM;
				else
					*AttribValues &= ~FILE_ATTRIBUTE_SYSTEM;
				break;

			default :
				return 1;
		}
	}

	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



DLLExports int WINAPI sift( LPTSTR inString )
{
	if ( !inString )
		return -1;


	FreeList();

	ParseArgs( inString, 5 );

	int Mode = MODE_UNKNOWN;
	int WhichStamp = STAMP_LAST_WRITE;
	DWORD AttribMask = FILE_ATTRIBUTE_DIRECTORY, AttribValues = 0;
	void *FirstRange = NULL;
	wchar_t ExcludeDescriptions[MAX_ARG_LEN] = L"";

	unsigned int n = 0, tempn = 0;
	LPTSTR inFilespec = NULL;
	LPTSTR ArrayName = NULL;
	int argtype[MAX_ARGS];
	int NumFilespecs = 0;
	int Recurse = 0;
	BOOL FullPaths = FALSE, Verbose = FALSE, Debug = FALSE;
	size_t LongestFilename = 0, LongestSize = 0, LongestStamp = 0;
	size_t len = 0;
	unsigned int DumpWidth = 0;
	int rv = 0;

	wchar_t CmdBuf[MAX_ARG_LEN] = L"";

	BOOL BeSmart = TRUE, UTC = FALSE, ISOFormat = FALSE, ShouldAddCommas = TRUE;
	wchar_t TimeSepString[SHORT_BUF_LEN] = L" ";
	wchar_t StampString[SHORT_BUF_LEN] = L"";


	for ( int i = 0; i < MAX_ARGS; i++ )
		argtype[i] = 0;

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcscmp( arg[i], L"/?" ) ) {
				ShowCmdHelp( SiftHelpText, FALSE );
				FreeLinkedList( &FirstRange );
				return 0;
			}
			else if ( !wcscmp( arg[i], L"/??" ) ) {
				ShowCmdHelp( SiftHelpText, TRUE );
				FreeLinkedList( &FirstRange );
				return 0;
			}
			else if ( ( Mode == MODE_UNKNOWN ) && ( !_wcsicmp( arg[i], L"LARGEST" ) || !_wcsicmp( arg[i], L"/L" ) || !_wcsicmp( arg[i], L"BIGGEST" ) || !_wcsicmp( arg[i], L"/B" ) ) )
				Mode = MODE_LARGEST;
			else if ( ( Mode == MODE_UNKNOWN ) && ( !_wcsicmp( arg[i], L"SMALLEST" ) || !_wcsicmp( arg[i], L"/S" ) ) )
				Mode = MODE_SMALLEST;
			else if ( ( Mode == MODE_UNKNOWN ) && ( !_wcsicmp( arg[i], L"OLDEST" ) || !_wcsicmp( arg[i], L"/O" ) ) )
				Mode = MODE_OLDEST;
			else if ( ( Mode == MODE_UNKNOWN ) && ( !_wcsicmp( arg[i], L"NEWEST" ) || !_wcsicmp( arg[i], L"/N" ) ) )
				Mode = MODE_NEWEST;
			else if ( !_wcsicmp( arg[i], L"/L" ) || !_wcsicmp( arg[i], L"/B" ) || !_wcsicmp( arg[i], L"/S" ) || !_wcsicmp( arg[i], L"/O" ) || !_wcsicmp( arg[i], L"/N" ) ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Too many criteria \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/R" ) ) {
				Recurse |= SIFT_RECURSIVE;
				FullPaths = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/T", 2 ) || !_wcsnicmp( arg[i], L"/D", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( towupper( arg[i][j] ) == 'A' )
					WhichStamp = STAMP_LAST_ACCESS;
				else if ( towupper( arg[i][j] ) == 'C' )
					WhichStamp = STAMP_CREATED;
				else if ( towupper( arg[i][j] ) == 'W' )
					WhichStamp = STAMP_LAST_WRITE;
				else {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown file stamp \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/A", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseAttribs( arg[i] + j, &AttribMask, &AttribValues ) ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad attributes \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/F" ) )
				FullPaths = TRUE;
			else if ( !_wcsicmp( arg[i], L"/I" ) )
				ISOFormat = TRUE;
			else if ( !_wcsicmp( arg[i], L"/U" ) )
				UTC = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/V", 2 ) ) {
				Verbose = TRUE;
				DumpWidth = 0;

				int j = 2;
				if ( arg[i][j] ) {
					if ( IsColonOrEquals( arg[i][j] ) )
						j++;

					if ( ParseInt( arg[i] + j, &DumpWidth, NULL ) ) {
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Bad width \"%s\"\r\n", arg[i] );
						goto ErrorExit;
					}
					else if ( ( DumpWidth != 0 ) && ( ( DumpWidth < MIN_USER_WIDTH ) || ( DumpWidth > MAX_USER_WIDTH ) ) ) {
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Width must be 0 or %u - %u\r\n", MIN_USER_WIDTH, MAX_USER_WIDTH );
						goto ErrorExit;
					}
				}
			}
			else if ( !_wcsicmp( arg[i], L"/Z" ) )
				BeSmart = FALSE;
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				while ( arg[i][j] ) {
					switch ( towupper( arg[i][j++] ) ) {
						case 'C' :
							ShouldAddCommas = FALSE;
							break;

						case 'D' :
							Recurse |= SIFT_NO_HIDDEN_DIRS;
							break;

						case 'I' :
							wcscpy_s( ExcludeDescriptions, MAX_ARG_LEN, L"%_dname" );
							ExpandVariables( ExcludeDescriptions, 1 );
							break;

						case 'J' :
							Recurse |= SIFT_NO_JUNCTIONS;
							break;

						default:
							_wcsupr_s( arg[i], MAX_ARG_LEN );
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Unknown option \"%s\"\r\n", arg[i] );
							goto ErrorExit;
					}
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/C",2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				wcscpy_s( TimeSepString, SHORT_BUF_LEN, arg[i] + j );
			}
			else if ( !_wcsicmp( arg[i], L"/$" ) )
				Debug = TRUE;
			else if ( !wcsncmp( arg[i], L"/[", 2 ) || !wcsncmp( arg[i], L"/![", 3 ) ) {
				if ( ParseRangeOption( arg[i], &FirstRange ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/I", 2 ) ) {
				if ( ParseDescriptionRange( arg[i] + 2, FALSE, TRUE, &FirstRange ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/!I", 3 ) ) {
				if ( ParseDescriptionRange( arg[i] + 3, TRUE, TRUE, &FirstRange ) )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else if ( !n && !ParseInt( arg[i], &tempn, NULL ) )
				n = tempn;
			else if ( !ArrayName ) {
				ArrayName = arg[i];
				if ( ArrayName[0] == '%' )
					if ( ArrayName[1] )
						ArrayName++;
			}
			else {
				argtype[i] = 1;
				NumFilespecs++;
			}
		}
	}

	if ( !ArrayName ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing array name\r\n" );
		goto ErrorExit;
	}

	if ( n > MAX_ITEMS ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Too many items (1 - %u)\r\n", MAX_ITEMS );
		goto ErrorExit;
	}
	else if ( !n )
		n = 10;

	if ( Mode == MODE_UNKNOWN )
		Mode = MODE_LARGEST;

	if ( ( Mode == MODE_LARGEST ) || ( Mode == MODE_SMALLEST ) )
		AttribMask |= FILE_ATTRIBUTE_DIRECTORY;

	_wcsupr_s( ArrayName, MAX_ARG_LEN );

	BOOL GoodName = TRUE;
	for ( int i = 0; ArrayName[i]; i++ ) {
		if ( !iswalnum( ArrayName[i] ) )
			if ( ( ArrayName[i] != '_' ) && ( ArrayName[i] != '$' ) )
				GoodName = FALSE;
	}
	if ( !iswalpha( ArrayName[0] ) )
		GoodName = FALSE;

	if ( !GoodName ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad array name \"%s\"\r\n", ArrayName );
		goto ErrorExit;
	}

	if ( NumFilespecs > 1 )
		FullPaths = TRUE;


	ScreenWidth = OutputWidth();
	NumItems = 0;


	if ( Debug ) {
		Printf( L"ArrayName  : %s\r\n", ArrayName );
		Printf( L"n          : %u\r\n", n );
		Printf( L"Mode       : %s\r\n", ModeName[Mode] );
		Printf( L"WhichStamp : %s\r\n", StampName[WhichStamp] );
		Printf( L"Recurse    : %s\r\n", BoolValue[Recurse & SIFT_RECURSIVE] );
		Printf( L"FullPaths  : %s\r\n", BoolValue[FullPaths] );
		Printf( L"ScreenWidth: %d\r\n", ScreenWidth );
		Printf( L"Verbose    : %s\r\n", BoolValue[Verbose] );
	}

	if ( NumFilespecs ) {
		for ( int i = 0; i < numargs; i++ )
			if ( argtype[i] )
				if ( SiftFiles2( arg[i], AttribMask, AttribValues, n, Mode, WhichStamp, Recurse, FullPaths, ExcludeDescriptions, FirstRange, Debug ) )
					rv = 2;
	}
	else if ( SiftFiles2( L"*", AttribMask, AttribValues, n, Mode, WhichStamp, Recurse, FullPaths, ExcludeDescriptions, FirstRange, Debug ) )
		rv = 2;


	//  create new array:

	wcscpy_s( CmdBuf, MAX_ARG_LEN, L" /Q " );
	wcscat_s( CmdBuf, MAX_ARG_LEN, ArrayName );
	CallInternal( L"unsetarray", CmdBuf );				//	UnsetArray_Cmd( CmdBuf );

	if ( NumItems ) {
		swprintf( CmdBuf, MAX_ARG_LEN, L" %s[%d,3]", ArrayName, NumItems );
		CallInternal( L"setarray", CmdBuf );			//	SetArray_Cmd( CmdBuf );
	}


	//  fill the array:

	FileEntry *Item = FirstItem;

	for ( int i = 0; i < NumItems; i++ ) {
		swprintf( CmdBuf, MAX_ARG_LEN, L"%s[%u,0]=%s", ArrayName, i, &(Item->Filename) );
		CallInternal( L"set", CmdBuf );						//	Set_Cmd( CmdBuf );
		len = wcslen( &(Item->Filename) );
		if ( len > LongestFilename )
			LongestFilename = len;

		swprintf( StampString, MAX_ARG_LEN, L"%I64u", Item->FileSize );
		if ( ShouldAddCommas )
			InsertCommas( StampString, SHORT_BUF_LEN );		//	AddCommas( StampString );
		swprintf( CmdBuf, MAX_ARG_LEN, L"%s[%u,1]=%s", ArrayName, i, StampString );
		CallInternal( L"set", CmdBuf );						//	Set_Cmd( CmdBuf );
		len = wcslen( StampString );
		if ( len > LongestSize )
			LongestSize = len;

		swprintf( CmdBuf, MAX_ARG_LEN, L"%s[%u,2]=", ArrayName, i );
		StampToString( StampString, Item->FileStamp, ( Item->VolumeFAT && ( WhichStamp == STAMP_CREATED ) ), ( Item->VolumeNTFS && BeSmart ), UTC, ISOFormat, TimeSepString );
		wcscat_s( CmdBuf, MAX_ARG_LEN, StampString );
		CallInternal( L"set", CmdBuf );						//	Set_Cmd( CmdBuf );
		len = wcslen( StampString );
		if ( len > LongestStamp )
			LongestStamp = len;

		Item = Item->Next;
	}

	FreeList();


	//  report and exit:

	if ( Verbose )
		DumpArray( ArrayName, LongestFilename, LongestSize, LongestStamp, WhichStamp, DumpWidth );

	Report( NumItems, AttribMask, AttribValues, Mode, ArrayName, NumItems );
	SiftFound = NumItems;
	FreeLinkedList( &FirstRange );
	return rv;


ErrorExit:

	FreeList();
	FreeLinkedList( &FirstRange );
	return 1;
}



DLLExports int WINAPI _sift( LPTSTR outString )
{
	if ( !outString )
		return -1;

	Sprintf( outString, L"%d", SiftFound );
	return 0;
}



DLLExports int WINAPI f_sift( LPTSTR String )
{
	if ( !String )
		return -1;


	FreeList();

	ParseArgs( String, 2 );


	int Mode = MODE_UNKNOWN;
	int WhichStamp = STAMP_LAST_WRITE;
	DWORD AttribMask = 0, AttribValues = 0;
	unsigned int Flags = 0;
	int Recurse = 0;
	wchar_t ExcludeDescriptions[MAX_ARG_LEN] = L"";

	
	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filespec\r\n" );
		return -1;
	}


	if ( arglen[1] ) {
		switch ( towupper( arg[1][0] ) ) {

			case 'L' :
			case 'B' :
				Mode = MODE_LARGEST;
				break;

			case 'S' :
				Mode = MODE_SMALLEST;
				break;

			case 'O' :
				Mode = MODE_OLDEST;
				break;

			case 'N' :
				Mode = MODE_NEWEST;
				break;

			default :
				_wcsupr_s( arg[1], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad criterion \"%s\"\r\n", arg[1] );
				return -1;
		}
	}
	else
		Mode = MODE_LARGEST;


	if ( arglen[2] ) {
		if ( ParseAttribs( arg[2], &AttribMask, &AttribValues ) ) {
			_wcsupr_s( arg[2], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad attributes \"%s\"\r\n", arg[2] );
			return -1;
		}
	}


	if ( arglen[3] ) {
		if ( ParseInt( arg[3], &Flags, NULL ) ) {
			_wcsupr_s( arg[3], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad flags \"%s\"\r\n", arg[3] );
			return -1;
		}
	}


	if ( arglen[4] ) {
		switch ( towupper( arg[4][0] ) ) {

			case 'A' :
				WhichStamp = STAMP_LAST_ACCESS;
				break;

			case 'C' :
				WhichStamp = STAMP_CREATED;
				break;

			case 'W' :
				WhichStamp = STAMP_LAST_WRITE;
				break;

			default :
				_wcsupr_s( arg[4], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown file stamp \"%s\"\r\n", arg[4] );
				return -1;
		}
	}


	Recurse = Flags & ( SIFT_RECURSIVE | SIFT_NO_JUNCTIONS | SIFT_NO_HIDDEN_DIRS );

	if ( ( Mode == MODE_LARGEST ) || ( Mode == MODE_SMALLEST ) )
		AttribMask |= FILE_ATTRIBUTE_DIRECTORY;

	if ( Flags & 0x0080 ) {
		wcscpy_s( ExcludeDescriptions, MAX_ARG_LEN, L"%_dname" );
		ExpandVariables( ExcludeDescriptions, 1 );
	}


	String[0] = '\0';

	if ( SiftFiles2( arg[0], AttribMask, AttribValues, 1, Mode, WhichStamp, Recurse, TRUE, ExcludeDescriptions, NULL, ( Flags & 0x0100 ) > 0 ) )
		return 0;

	if ( NumItems == 0 )
		return 0;

	wcscpy_s( String, MAX_FILENAME_LEN, &(FirstItem->Filename) );
	FreeList();
	return 0;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/sift.html";
	piInfo.pszDescription	= (LPTSTR) L"File sorting plugin for TCC";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	if ( UnsupportedVersion() ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"TCC v10.00 or higher required!\r\n" );
		return -1;
	}

	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%d.%d.%d loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	FreeList();
	return 0;
}



LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}



DLLExports int WINAPI Help( LPTSTR inString )
{
	if ( !_wcsicmp( inString, L"sift" ) ) {
		MessageBox( GetForegroundWindow(), SiftHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"_sift" ) ) {
		MessageBox( GetForegroundWindow(), _SiftHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else if ( !_wcsicmp( inString, L"@sift" ) ) {
		MessageBox( GetForegroundWindow(), f_SiftHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else
		return 0;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( !_wcsicmp( inString, L"sift" ) ) {
		wcscpy_s( outString, SHORT_BUF_LEN, UsageText( SiftHelpText ) );
		return 1;
	}
	else
		return 0;
}
